# For any of the general bots, the test cases are generated by the following:
# First, loop through all the kinematics of the robot (h and p matrices) in the test_cases folder
# For each of these, we generate 20 random positions for the robot to be in (q values)
# We run forward kinematics for each of these, then run inverse kinematics on the result
# We then run forward kinematics on the result to make sure it is the same as the forward kinematics of the input

# For general bots, we don't have access to forward kinematics, so we can't do the same test
# Instead, we will just run inverse kinematics on all of the given test cases in the test_cases folder and make sure that we actually return a nonzero result
# THIS IS NOT A TEST FOR CORRECTNESS ON HARD-CODED BOTS, ONLY ON GENERAL BOTS

# Requires numpy to be installed

# Import the test case csv files

from collections import namedtuple
import numpy as np
import unittest
from ik_geo import Robot
from math import pi

# Import the csv files
import csv
import os

TestBot = namedtuple("TestBot", ["casename", "robot", "testcases"])
TestCaseGeneral = namedtuple("TestCaseGeneral", ["hVals", "pVals"])
TestCaseHardCoded = namedtuple(
    "TestCaseHardCoded", ["rotationMatrix", "positionVector"]
)


generalFilenames = [
    "IK_2_intersecting",
    "IK_2_parallel",
    "IK_3_parallel_2_intersecting",
    "IK_3_parallel",
    "IK_spherical",
    "IK_spherical_2_parallel",
    "IK_spherical_2_intersecting",
    "IK_gen_6_dof",
]
hardcodedFilenames = [
    "IRB_6640",
    "spherical_bot",
    "three_parallel_bot",
    "two_parallel_bot",
    "ur5"
]

epsilon = 1e-3


# Test the general robots
class TestGeneralRobots(unittest.TestCase):
    def check_general_robot(self, bot):
        np.random.seed(0)
        for configNum, testCase in enumerate(bot.testcases):
            hMatrix = np.reshape(testCase.hVals, (6, 3))
            pMatrix = np.reshape(testCase.pVals, (7, 3))
            robot = bot.robot(hMatrix, pMatrix)

            # Generate 20 random robot configurations
            qVals = np.random.rand(20, 6) * 2 * pi - pi
            # Generate fewer for the 6 dof bot because it is slower
            if bot.casename == "IK_gen_6_dof":
                qVals = np.random.rand(5, 6) * 2 * pi - pi
            for i, q in enumerate(qVals):
                # Get the forward kinematics result and then run inverse to see if we get the same thing
                forward_kinematics = robot.forward_kinematics(q)
                rotation = np.array(forward_kinematics[0])
                translation = np.array(forward_kinematics[1])

                # Get the inverse kinematics
                results = robot.get_ik_sorted(
                    forward_kinematics[0], forward_kinematics[1]
                )

                self.assertFalse(
                    len(results) == 0,
                    msg=f"Failed on test {i + 1} of {bot.casename} configuration {configNum + 1} with \nqVals: {str(q)} \
                                    \nRotation matrix expected: \n{str(rotation)}\nTranslation expected: {str(translation)} \
                                    \nhMatrix: \n{str(hMatrix)}\npMatrix: \n{str(pMatrix)} \
                                    \nNo result was returned.",
                )
                result = results[0]

                # Run forward kinematics on the result to make sure it is the same as the input
                resultForward = robot.forward_kinematics(result[0])
                resultRotation = np.array(resultForward[0])
                resultTranslation = np.array(resultForward[1])

                # Check to make sure each value is roughly equal, up to 2pi
                for resultVal, expectedVal in zip(
                    rotation.flatten(), resultRotation.flatten()
                ):
                    self.assertAlmostEqual(
                        resultVal,
                        expectedVal,
                        delta=epsilon,
                        msg=f"Failed on test {i + 1} of {bot.casename} configuration {configNum + 1} with \nqVals: {str(q)}\n and result: {str(result[0])} \
                        \nRotation matrix expected: \n{str(rotation)}\nGot: \n{str(resultRotation)} \
                        \nTranslation expected: {str(translation)}\nGot: {str(resultTranslation)} \
                        \nhMatrix: \n{str(hMatrix)}\npMatrix: \n{str(pMatrix)}",
                    )
                for resultVal, expectedVal in zip(translation, resultTranslation):
                    self.assertAlmostEqual(
                        resultVal,
                        expectedVal,
                        delta=epsilon,
                        msg=f"Failed on test {i + 1} of {bot.casename} configuration {configNum + 1} with \nqVals: {str(q)}\n and result: {str(result[0])} \
                                            \nRotation matrix expected: \n{str(rotation)}\nGot: \n{str(resultRotation)} \
                                            \nTranslation expected: {str(translation)}\nGot: {str(resultTranslation)} \
                                            \nhMatrix: \n{str(hMatrix)}\npMatrix: \n{str(pMatrix)}",
                    )


def add_general_test(filename):
    # Setup the robot from the filename
    robot_name = (
        filename.replace("IK_", "")
        .replace("2", "two")
        .replace("3", "three")
        .replace("6", "six")
    )

    # Switch on robot name
    robot_instantiater = getattr(Robot, robot_name)

    testcases = []
    # Open ../../test_cases/FILENAME.csv
    with open(
        os.path.join(os.path.dirname(__file__), "..", "test_cases", filename + ".csv")
    ) as f:
        reader = csv.reader(f)
        # Skip the header
        next(reader)
        for row in reader:
            if not row:
                continue
            # Parse the given h and p vals to get a valid robot configuration
            hVals = [float(x) for x in row[0:18]]
            pVals = [float(x) for x in row[18:39]]
            testcases.append(TestCaseGeneral(hVals, pVals))
    bot = TestBot(filename, robot_instantiater, testcases)
    setattr(
        TestGeneralRobots,
        "test_" + filename,
        lambda self: self.check_general_robot(bot),
    )


class TestHardcodedBots(unittest.TestCase):
    def check_hardcoded_robot(self, bot):
        for testCase in bot.testcases:
            rotationMatrix = np.array(
                [
                    testCase.rotationMatrix[i : i + 3]
                    for i in range(0, len(testCase.rotationMatrix), 3)
                ]
            )
            positionVector = testCase.positionVector
            result = bot.robot.get_ik(rotationMatrix, positionVector)
            self.assertFalse(
                len(result) == 0,
                msg=f"Failed on {bot.casename} with rotation matrix: \n{rotationMatrix} and position vector: \n{positionVector}. \nNo result was returned.",
            )


def add_hardcoded_test(filename):
    # Setup the robot from the filename
    robot = getattr(Robot, filename.replace("IRB_6640", "irb6640"))()

    testcases = []
    # Open ../../test_cases/FILENAME.csv
    with open(
        os.path.join(os.path.dirname(__file__), "..", "test_cases", filename + ".csv")
    ) as f:
        reader = csv.reader(f)
        # Skip the header
        next(reader)
        for row in reader:
            if not row:
                continue
            # Parse the given h and p vals to get a valid robot configuration
            rotationMatrix = [float(x) for x in row[0:9]]
            positionVector = [float(x) for x in row[9:12]]
            testcases.append(TestCaseHardCoded(rotationMatrix, positionVector))
    bot = TestBot(filename, robot, testcases)
    setattr(
        TestHardcodedBots,
        "test_" + filename,
        lambda self: self.check_hardcoded_robot(bot),
    )


if __name__ == "__main__":
    # Generate classes for each general test case
    for filename in generalFilenames:
        add_general_test(filename)
    for filename in hardcodedFilenames:
        add_hardcoded_test(filename)
    unittest.main()



from ik_geo import Robot

# Sample assumes you have numpy installed globally
import numpy as np


def run_ik_hardcoded():
    print("\nRunning hardcoded inverse kinematics:\n-----------------------------")
    # Create the robot, type ur5
    # This can be done with string literals as well as factory methods (shown below)
    robot = Robot.ur5()
    # Get the inverse kinematics
    # The first argument is the rotation matrix (3x3, row major)
    # The second argument is the position vector (3x1)
    solutions = robot.get_ik(
        [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]], [0.0, 0.0, 0.0]
    )
    for (q, ls) in solutions:
        print("Soution: (" + ("Least Squares" if ls else "Exact")+ ")")
        for qVal in q:
            print(qVal)
        print("-----------------------------")
    print("-----------------------------")


def run_ik_general():
    print("\nRunning general inverse kinematics:\n-----------------------------")

    # Create the kinematics
    # Paremeters are deconstructed h matrix and p matrix
    # I'm using the code from the Irb6640 to for real kinematics
    zv = [0.0, 0.0, 0.0]
    ex = [1.0, 0.0, 0.0]
    ey = [0.0, 1.0, 0.0]
    ez = [0.0, 0.0, 1.0]
    hMat = np.array([ez, ey, ey, ex, ey, ex])
    pMat = np.array(
        [zv, [0.32, 0, 0.78], [0, 0, 1.075], [1.1425, 0, 0.2], ez, ez, [0.2, 0, 0]]
    )

    # MUST SET THE KINEMATICS OBJECT BEFORE RUNNING IK IN GENERAL CASE
    robot = Robot.spherical_two_intersecting(hMat, pMat)

    # Get the inverse kinematics
    # The first argument is the rotation matrix (3x3, row major deconstructed)
    # The second argument is the position vector (3x1)
    solutions = robot.get_ik_sorted([[0, 0, 1], [0, 1, 0], [-1, 0, 0]], [-1, 3, 0])

    for (q, error, ls) in solutions:
        print("Soution: (" + ((f"Least Squares, error={error}") if ls else "Exact")+ ")")
        for qVal in q:
            print(qVal)
        print("-----------------------------")
    print("-----------------------------")

run_ik_hardcoded()
run_ik_general()



from typing import List, Tuple, Annotated, Literal
import numpy as np
from numpy.typing import NDArray

class Robot:
    """
    Representation of the robot for inverse kinematics

    """

    def get_ik(
        self,
        R: (
            Annotated[NDArray[np.generic], Literal[3, 3]]
            | Annotated[List[List[float]], [3, 3]]
        ),
        t: Annotated[NDArray[np.generic], Literal[3]] | Annotated[List[float], [3]],
    ) -> List[Tuple[List[float], bool]]:
        """
        Compute the inverse kinematics solutions for the robot.

        Args:
            R: The rotation matrix to use for the inverse kinematics
            t: The position vector to use for the inverse kinematics
        Returns:
            A list of tuples containing solutions.
            A solution contains the rotation values of each joint and whether the solution is least squares
        """

    def get_ik_sorted(
        self,
        R: (
            Annotated[NDArray[np.generic], Literal[3, 3]]
            | Annotated[List[List[float]], [3, 3]]
        ),
        t: Annotated[NDArray[np.generic], Literal[3]] | Annotated[List[float], [3]],
    ) -> List[Tuple[List[float], float, bool]]:
        """
        Compute the inverse kinematics solutions for the robot, as well as the error of each solution, sorted by error.

        Args:
            R: The rotation matrix to use for the inverse kinematics
            t: The position vector to use for the inverse kinematics
        Returns:
            A list of tuples containing solutions.
            A solution contains the rotation values of each joint, the error of the solution, and whether the solution is least squares
                They are sorted by error from least to greatest
        """

    def forward_kinematics(
        self,
        q: Annotated[List[float], [6]] | Annotated[NDArray[np.generic], Literal[6]],
    ) -> Tuple[List[List[float]], List[float]]:
        """
        Get the forward kinematics for the robot, not implemented for hardcoded bots

        Args:
            q: The rotation values of each joint
        Returns:
            A tuple containing the rotation matrix and position vector
        """

    @classmethod
    def irb6640(cls) -> "Robot": ...

    # @classmethod
    # def kuka_r800_fixed_q3(cls) -> "Robot":
    #     ...

    @classmethod
    def ur5(cls) -> "Robot": ...
    @classmethod
    def three_parallel_bot(cls) -> "Robot": ...
    @classmethod
    def two_parallel_bot(cls) -> "Robot": ...
    @classmethod
    def spherical_bot(cls) -> "Robot": ...

    # @classmethod
    # def rrc_fixed_q6(cls) -> "Robot":
    #     return cls("RrcFixedQ6")

    # @classmethod
    # def yumi_fixed_q3(cls) -> "Robot":
    #     return cls("YumiFixedQ3")

    @classmethod
    def spherical_two_intersecting(
        cls,
        h: (
            Annotated[NDArray[np.generic], Literal[6, 3]]
            | Annotated[List[List[float]], [6, 3]]
        ),
        p: (
            Annotated[NDArray[np.generic], Literal[7, 3]]
            | Annotated[List[List[float]], [7, 3]]
        ),
    ) -> "Robot":
        ...

    @classmethod
    def spherical_two_parallel(
        cls,
        h: (
            Annotated[NDArray[np.generic], Literal[6, 3]]
            | Annotated[List[List[float]], [6, 3]]
        ),
        p: (
            Annotated[NDArray[np.generic], Literal[7, 3]]
            | Annotated[List[List[float]], [7, 3]]
        ),
    ) -> "Robot":
        ...

    @classmethod
    def spherical(
        cls,
        h: (
            Annotated[NDArray[np.generic], Literal[6, 3]]
            | Annotated[List[List[float]], [6, 3]]
        ),
        p: (
            Annotated[NDArray[np.generic], Literal[7, 3]]
            | Annotated[List[List[float]], [7, 3]]
        ),
    ) -> "Robot":
        ...

    @classmethod
    def three_parallel_two_intersecting(
        cls,
        h: (
            Annotated[NDArray[np.generic], Literal[6, 3]]
            | Annotated[List[List[float]], [6, 3]]
        ),
        p: (
            Annotated[NDArray[np.generic], Literal[7, 3]]
            | Annotated[List[List[float]], [7, 3]]
        ),
    ) -> "Robot":
        ...

    @classmethod
    def three_parallel(
        cls,
        h: (
            Annotated[NDArray[np.generic], Literal[6, 3]]
            | Annotated[List[List[float]], [6, 3]]
        ),
        p: (
            Annotated[NDArray[np.generic], Literal[7, 3]]
            | Annotated[List[List[float]], [7, 3]]
        ),
    ) -> "Robot":
        ...

    @classmethod
    def two_parallel(
        cls,
        h: (
            Annotated[NDArray[np.generic], Literal[6, 3]]
            | Annotated[List[List[float]], [6, 3]]
        ),
        p: (
            Annotated[NDArray[np.generic], Literal[7, 3]]
            | Annotated[List[List[float]], [7, 3]]
        ),
    ) -> "Robot":
        ...

    @classmethod
    def two_intersecting(
        cls,
        h: (
            Annotated[NDArray[np.generic], Literal[6, 3]]
            | Annotated[List[List[float]], [6, 3]]
        ),
        p: (
            Annotated[NDArray[np.generic], Literal[7, 3]]
            | Annotated[List[List[float]], [7, 3]]
        ),
    ) -> "Robot":
        ...

    @classmethod
    def gen_six_dof(
        cls,
        h: (
            Annotated[NDArray[np.generic], Literal[6, 3]]
            | Annotated[List[List[float]], [6, 3]]
        ),
        p: (
            Annotated[NDArray[np.generic], Literal[7, 3]]
            | Annotated[List[List[float]], [7, 3]]
        ),
    ) -> "Robot":
        ...

